/*
  ==================================================================================================
  _____ _____          _      ____   _____  _____ _______ _____  _      
 |  __ \_   _|   /\   | |    / __ \ / ____|/ ____|__   __|  __ \| |     
 | |  | || |    /  \  | |   | |  | | |  __| |       | |  | |__) | |     
 | |  | || |   / /\ \ | |   | |  | | | |_ | |       | |  |  _  /| |     
 | |__| || |_ / ____ \| |___| |__| | |__| | |____   | |  | | \ \| |____ 
 |_____/_____/_/    \_\______\____/ \_____|\_____|  |_|  |_|  \_\______|
                                                                        

  Manages all the other objects, holds state related to the user interaction with the view
  ==================================================================================================
*/
var DialogCtrl = function() {
  'use strict';
  
  var self;
  var analytics;
  var inkPathsHandler;
  var ui;
  var nbInterface;
  var kappInterface;

  function DialogCtrl(pAnalytics, pInkPathsHandler, pUi, pNbInterface, pKappInterface) {
    // Rename to put the objects in scope of DialogCtrl's functions
    self = this;
    // Set these again for scoping issues
    analytics = pAnalytics;
    inkPathsHandler = pInkPathsHandler;
    ui = pUi;
    nbInterface = pNbInterface;
    kappInterface = pKappInterface;

    ui.status.toggleMessage(true, ui.MESSAGE_LOADING);
    this.crop = new DialogCtrlCrop(ui, this);
    this.scrubber = new DialogCtrlScrub(ui, this);
    this.ink = new DialogCtrlInk(ui, nbInterface, this);
    this.board = new DialogCtrlBoard(inkPathsHandler, ui, nbInterface, kappInterface, this);
    this.initEventHandlers();
    this.initNbCommunication();
  }

  DialogCtrl.prototype.initNbCommunication = function() {
    var nbListeners = [
      ["initialize", self.board.initState],
      ["update", self.board.updateState],
      ["fallback", self.board.fallbackInit],
      ["readyToDraw", function() {
        nbInterface.inkToPage(BoardState, self.ink.drawData);
      }],
      ["seek", self.scrubber.seekTo],
      ["clear", self.board.clearBoard],
      ["disconnect", self.board.disconnected]
    ];
    var nbReadyCallback = function() {
      self.board.sendWave();
    };
    nbInterface.onNbReady(nbListeners, nbReadyCallback);
  };

  DialogCtrl.prototype.initEventHandlers = function() {
    // Prevent scrolling
    window.onscroll = function () {
      window.scrollTo(0,0);
    };

    /*
    ------------
      SCRUBBER
    ------------
    */
    // Make the slider trigger change on every input
    ui.elems.$scrubber.on('input', function() {
      self.scrubber.changeScrubber(this.value);
    // Everytime the scrubber changes, update the self.messageIndex and redraw the canvas
    }).change(function() {
      self.scrubber.changeScrubber(this.value);
    // Analytics tracking once per scrubber drag/click
    }).mouseup(function() {
      analytics.trackEvent('button', 'click', 'scrubber drag', 1);
    });

    // Clicking will step once. If mouse is held down for longer, then it will accelerate
    // to do multiple steps (so it essentially becomes a fast forward/backward)
    ui.elems.$stepBack.mousedown(function () {
      analytics.trackEvent('button', 'click', 'step back', 1);
      self.scrubber.startAcceleratedStepBack();
    }).mouseup(function() {
      self.scrubber.stepScrubTime(-1);
    });
    ui.elems.$stepForward.mousedown(function () {
      analytics.trackEvent('button', 'click', 'step forward', 1);
      self.scrubber.startAcceleratedStepForward();
    }).mouseup(function() {
      self.scrubber.stepScrubTime(1);
    });

    // Stop the fast forward/backward from the step buttons
    ui.elems.$doc.mouseup(function() {
      self.scrubber.clearAccelStepIntervals();
    });

    // Jump to prev snapshot tick
    ui.elems.$skipBack.click(function () {
      analytics.trackEvent('button', 'click', 'skip back', 1);
      self.scrubber.seekNearestSnapshotTick(-1);
    });

    // Jump to next snapshot tick
    ui.elems.$skipForward.click(function () {
      analytics.trackEvent('button', 'click', 'step forward', 1);
      self.scrubber.seekNearestSnapshotTick(1);
    });

    // Play, Pause, or Restart playing at beginning, which occurs depends on curr state of scrubber
    ui.elems.$playPause.click(function() {
      self.scrubber.playPause(
        //Analytics callbacks
        function() { analytics.trackEvent('button', 'click', 'pause', 1); },
        function() { analytics.trackEvent('button', 'click', 'play', 1); }, 
        function() { analytics.trackEvent('button', 'click', 'restart', 1); }
      );
    });

    /*
    -----------
      CROPPER
    -----------
    */
    // This only gets triggered on drag for some reason.
    // Need to additionally update manually on mousedowns.
    ui.elems.$doc.mousemove(function(event){
      self.crop.updateCursor(event);
    });
    // Crop button toggles cropping mode.
    ui.elems.$cropClickable.click(function() {
      analytics.trackEvent('button', 'click', 'cropper toggle', 1);
      self.crop.toggleCropping(!self.crop.isCropping);
    });
    // When click and drag over the canvas, it will bring up a cropping selection
    ui.elems.$curtainContDivs.mousedown(function(event) {
      self.crop.updateDraggingSelection(event);
    });
    // When click and drag over the cropping selection, it will move
    ui.elems.$cropperBox.mousedown(function(event) {
      self.crop.updateMovingSelection(event);
    });
    // Clicking without dragging in the cropper box will ink to the current page
    ui.elems.$cropperBox.mouseup(function(event) {
      self.crop.checkInkToPage(function() {
        //Analytics callback
        analytics.trackEvent('button', 'click', 'ink to page - cropper', 1);
      });
    });
    // Handles mouse clicking up while cropping
    //  clicking up in the crop curtain will exit cropping mode
    //  clicking up from a drag over the canvas will stop selecting the cropping box
    //  clicking up from a drag over the cropping selection will stop moving the cropping box
    ui.elems.$doc.mouseup(
      self.crop.checkEndMoveSelectCrop
    );

    /*
    ----------
      INKING
    ----------
    */
    ui.elems.$animatedButtonCont.click(
      ui.buttonClickAnimation
    );
    // Inserts ink into the page
    ui.elems.$insertInkClickable.click(function() {
      analytics.trackEvent('button', 'click', 'ink to page', 1);
      self.ink.insertToPage();
    });
    // Creates a new page and inserts ink into it
    ui.elems.$insertInkNewPageClickable.click(function() {
      analytics.trackEvent('button', 'click', 'ink to new page', 1);
      self.ink.insertToNewPage();
    });
    // Changes the selected ink color
    ui.elems.$blackInk.click(function() {
      analytics.trackEvent('button', 'click', 'black', 1);
      self.ink.activateColorBlack($(this));
    });
    ui.elems.$redInk.click(function () {
      analytics.trackEvent('button', 'click', 'red', 1);
      self.ink.activateColorRed($(this));
    });
    ui.elems.$blueInk.click(function() {
      analytics.trackEvent('button', 'click', 'blue', 1);
      self.ink.activateColorBlue($(this));
    });
    ui.elems.$greenInk.click(function() {
      analytics.trackEvent('button', 'click', 'green', 1);
      self.ink.activateColorGreen($(this));
    });

    /*
    -------------
      HELP INIT
    -------------
    */
    // Clicking on the help icon will toggle the help mode
    ui.elems.$helpCont.click(function() {
      analytics.trackEvent('button', 'click', 'help', 1);
      ui.help.toggleHelp(!ui.help.isHelp);
    });
    // Clicking anywhere in the dialog while help is active will advance to the next instructions
    $(document).click(function() {
      ui.help.advanceHelpIfActive();
    });
  };

  return DialogCtrl;
}();