/*
  ==================================================================================================
   ___ ___  ___  ___ 
  / __| _ \/ _ \| _ \
 | (__|   / (_) |  _/
  \___|_|_\\___/|_|  
  
  DialogCtrlCrop maintains state and provides functionality for cropping      
  ==================================================================================================
  Cropper Tool
    cursorX - the current x position of the mouse relative to the canvas
    cursorY - the current y position of the mouse relative to the canvas
    cropStart - the starting point of the crop selection
    cropEnd - the end point of the crop selection
    moveStart - the starting point of the crop move
    moveEnd - the ending point of the crop move
    isSelecting - when the user is drag selecting a crop box
    isMoving - when the user is drag moving the crop box
    isCropping - when the user has a crop box drawn
    mouseMoveInterval - throttles the number of times mousemove performs functions
    statusIsLive - whether the status was live before the last time we started cropping
    colorIsActive - whether the color controls were live before the last time we started cropping
*/
var DialogCtrlCrop = function() {
  'use strict';
  
  var TIME_DELAY = 42; //the rate at which mousemove and scrubber events are fired
  var CROPPING_THRESHOLD = 10; //the minimum size of the cropper box allowed

  var self;
  var ui;
  var dialogCtrl;

  function DialogCtrlCrop(pUi, pDialogCtrl) {
    self = this;
    ui = pUi;
    dialogCtrl = pDialogCtrl;
    this.cursorX = undefined;
    this.cursorY = undefined;
    this.cropStart = {x: 0, y:0 };
    this.cropEnd = {x: 0, y:0 };
    this.moveStart = {x: 0, y: 0};
    this.moveEnd = {x: 0, y: 0};
    this.isSelecting = false;
    this.isMoving = false;
    this.isCropping = false;
    this.mouseMoveInterval = undefined;
    this.statusIsLive = false;
    this.colorIsActive = false;
  }


  // Updates mouse cursor information
  DialogCtrlCrop.prototype.updateCursor = function(event) {
    this.cursorX = event.pageX - ui.elems.$curtainCont.offset().left;
    this.cursorY = event.pageY - ui.elems.$curtainCont.offset().top;
  };

  DialogCtrlCrop.prototype.updateDraggingSelection = function(event) {
    if (!self.isCropping) {
      return;
    }
    self.updateCursor(event);
    self.cropStart.x = self.cursorX;
    self.cropStart.y = self.cursorY;
    self.cropEnd.x = self.cropStart.x;
    self.cropEnd.y = self.cropStart.y;
    self.isSelecting = true;
    // Limit the amount of mousemove calls to reduce lag
    if (!self.mouseMoveInterval) {
      self.mouseMoveInterval = setInterval(self.onMouseMove, TIME_DELAY);
    }
  };

  DialogCtrlCrop.prototype.updateMovingSelection = function(event) {
    if (!self.isCropping) {
      return;
    }
    self.updateCursor(event);
    self.moveStart.x = self.cursorX;
    self.moveStart.y = self.cursorY;
    self.moveEnd.x = self.moveStart.x;
    self.moveEnd.y = self.moveStart.y;
    self.isMoving = true;
    // Limit the amount of mousemove calls to reduce lag
    if (!self.mouseMoveInterval) {
      self.mouseMoveInterval = setInterval(self.onMouseMove, TIME_DELAY);
    }
  };

  // Calculates the x and y difference from start to end, and the magnitude
  DialogCtrlCrop.prototype.getVector = function(start, end) {
    var vector = {};
    vector.x = end.x - start.x;
    vector.y = end.y - start.y;
    vector.magnitude = Math.sqrt(vector.x * vector.x + vector.y * vector.y);
    return vector;
  };

  // Toggles the dimming of the curtain, the cropping state, and the live status.
  // When set to false, the cropper start is reset
  DialogCtrlCrop.prototype.toggleCropping = function(state) {
    self.isCropping = state;
    if (self.isCropping) {
      //save for when we will toggle off cropping
      self.statusIsLive = ui.elems.$liveStatus.css("opacity") == 1;
      self.colorIsActive = ui.elems.$colorControls.css("opacity") == 1;
    } else {
      self.resetCropper();
    }
    ui.toggleCropping(self.isCropping, self.statusIsLive, self.colorIsActive);
  };

  // Handles mouse moving while cropping
  DialogCtrlCrop.prototype.onMouseMove = function() {
    if (self.isSelecting) {
      // Creating the selection area
      self.cropEnd.x = self.cursorX;
      self.cropEnd.y = self.cursorY;
      // Redraw the cropper window with the updated cropper box state
      ui.crop.drawCropperWindow(self.cropStart, self.cropEnd);
    } else if (self.isMoving) {
      // Moving the selection area
      self.moveEnd.x = self.cursorX;
      self.moveEnd.y = self.cursorY;
      // Get the distance moved and move the box accordingly
      var vector = self.getVector(self.moveStart, self.moveEnd);
      var ncropStart = {};
      var ncropEnd = {};
      ncropStart.x = self.cropStart.x + vector.x;
      ncropStart.y = self.cropStart.y + vector.y;
      ncropEnd.x = self.cropEnd.x + vector.x;
      ncropEnd.y = self.cropEnd.y + vector.y;
      ui.crop.drawCropperWindow(ncropStart, ncropEnd);
    } 
  };

  // Clicking without dragging in the cropper box will ink to the current page
  DialogCtrlCrop.prototype.checkInkToPage = function(analyticsCallback) {
    if (!self.isMoving) {
      return;
    }
    self.isSelecting = false;
    // See if there was a click with no drag
    var vector = self.getVector(self.moveStart, self.moveEnd);
    if (vector.magnitude <= 0) {
      dialogCtrl.ink.insertToPage();
      analyticsCallback();
    }
  };

  // Stops the cropper box selection if it was being dragged to select
  // Turns off cropping when the cropper curtain is clicked on
  // Stops the cropper box movement if it was being moved
  DialogCtrlCrop.prototype.checkEndMoveSelectCrop = function() {
    if (ui.help.isHelp) {
      return;
    }
    if (self.isSelecting) {
      // Done creating the selection area
      self.isSelecting = false;
      // If the selected area is too small, exit out of cropping
      // Also as a sideffect means if you click outside of the cropper box it will turn off cropping
      var vector = self.getVector(self.cropStart, self.cropEnd);
      if (vector.magnitude < CROPPING_THRESHOLD) {
        self.toggleCropping(false);
      }
    } else if (self.isMoving) {
      // Done moving the selection
      self.isMoving = false;
      // Update the cropper box state
      var vector = self.getVector(self.moveStart, self.moveEnd);
      self.cropStart.x += vector.x;
      self.cropStart.y += vector.y;
      self.cropEnd.x += vector.x;
      self.cropEnd.y += vector.y;
    }
    if (self.mouseMoveInterval) { 
      clearInterval(self.mouseMoveInterval); 
      self.mouseMoveInterval = undefined; 
    }
  };

  // Resets the cropper state
  DialogCtrlCrop.prototype.resetCropper = function() {
    this.cropStart.x = this.cropStart.y = this.cropEnd.x = this.cropEnd.y = 0;
    ui.crop.drawCropperWindow(this.cropStart, this.cropEnd);
  };

  return DialogCtrlCrop;
}();

  
