/*
  ==================================================================================================
  ___  ___ ___ _   _ ___ ___ ___ ___ 
 / __|/ __| _ \ | | | _ ) _ ) __| _ \
 \__ \ (__|   / |_| | _ \ _ \ _||   /
 |___/\___|_|_\\___/|___/___/___|_|_\  

  DialogCtrlScrub maintains state and provides functionality for scrubbing   
  ==================================================================================================
  Scrubber Timeline
    playInterval - interval to advance the timeline with play
    incrementBackInterval - acceleration of playback backwards with step
    incrementForwardInterval - acceleration of playback forwards with step
    stepBackInterval - playback backwards with step
    stepForwardInterval - playback forwards with step
    stepTimeout - delay when holding down step buttons before playback
    isCanvasUpdating - when the canvas update interval is queued up
    playing - when the play button is clicked and the session is playing through
    indexToDraw - used by the canvas drawing interval                                
*/
var DialogCtrlScrub = function() {
  'use strict';
  
  var PLAYBACK_DELAY = 100; //how fast to playback the timeline
  var ACCELERATION_DELAY = 500; //how fast to accelerate the playback
  var TIME_DELAY = 42; // the rate at which scrubber events are fired
  var PLAYBACK_START_DELAY = 200; //how long before the step buttons start playback

  var self;
  var ui;
  var dialogCtrl;

  function DialogCtrlScrub(pUi, pDialogCtrl) {
    self = this;
    ui = pUi;
    dialogCtrl = pDialogCtrl;
    this.incrementBackInterval = [];
    this.incrementForwardInterval = [];
    this.stepBackInterval = [];
    this.stepForwardInterval = [];
    this.stepTimeout = [];
    this.isCanvasUpdating = false;
    this.playing = false;
    this.playInterval = undefined;
    this.indexToDraw = -1;
    this.sliderTicks = [];
    this.activeTicks = [];
  }

  DialogCtrlScrub.prototype.initSlider = function() {
    ui.elems.$scrubber[0].min = 0;
    ui.elems.$scrubber[0].max = 0;
    this.updateSlider();
  };


  // Updates the slider ticks
  // Relies on the Boardstate being up to date
  DialogCtrlScrub.prototype.updateSlider = function() {
    ui.elems.$scrubber[0].max = BoardState.messageCount;
    this.calculateSliderTicks();
    this.addSliderTicks();
  };

  DialogCtrlScrub.prototype.changeScrubber = function(value) {
    if (!self.playing || (dialogCtrl.board.messageIndex == BoardState.messageCount && value < BoardState.messageCount) || 
        value == BoardState.messageCount) { 
      self.pause(); 
    }
    dialogCtrl.board.messageIndex = parseInt(value);
    // Start a polling interval - only actually draw every X ms passes
    self.indexToDraw = dialogCtrl.board.messageIndex;
    if (!self.isCanvasUpdating) {
      self.isCanvasUpdating = true;
      var canvasInterval = setInterval(function() {
        if (self.indexToDraw != -1) {
          dialogCtrl.board.updateCanvas(0, self.indexToDraw);
          self.indexToDraw = -1;
        } else {
          clearInterval(canvasInterval);
          self.isCanvasUpdating = false;
        }
      }, self.TIME_DELAY);
    }
  };

  DialogCtrlScrub.prototype.play = function() {
    ui.elems.$playPause[0].src = "img/pause.png";
    if (!self.playing) {
      self.playing = true;
      printDebug("starting playing interval");
      self.playInterval = setInterval(function() {
        self.stepScrubTime(1);
      }, PLAYBACK_DELAY);
    }
  };

  DialogCtrlScrub.prototype.pause = function() {
    ui.elems.$playPause[0].src = "img/play.png";
    self.playing = false;
    printDebug("pausing");
    clearInterval(self.playInterval);
  };

  DialogCtrlScrub.prototype.restart = function() { 
    var start = 0;
    self.setScrubTime(start);
    printDebug("restart");
    this.play();
  };

  // Increments/decrements the time on the scrubber - triggers input change
  DialogCtrlScrub.prototype.stepScrubTime = function(amount) {
    self.setScrubTime(parseInt(ui.elems.$scrubber[0].value) + amount);
  };

  // Set the time on the scrubber - triggers input event
  DialogCtrlScrub.prototype.setScrubTime = function(time) {
    var scrubber = ui.elems.$scrubber[0];
    time = time > scrubber.max ? scrubber.max : time < scrubber.min ? scrubber.min : time;
    scrubber.value = time;
    this.changeScrubber(time);
  };

  // Calculates the positions of the slider ticks and puts them into this.sliderTicks array
  DialogCtrlScrub.prototype.calculateSliderTicks = function() {
    var slider = ui.elems.$scrubber[0];
    var width = ui.elems.$scrubberCont.width();
    var max =  slider.max;    
    var min =  slider.min;
    // The offsets are magic numbers adjusted to notebook
    var offset_left = 7 / width * 100;
    var offset_right = 22 / width * 100;
    var steps = max - min;    
    var spacing = steps === 0 ? 0 : (100 - offset_left - offset_right) / steps;
    this.sliderTicks = [];
    for (var i = 0; i <= max - min ; i++) {
      this.sliderTicks[i] = (spacing * i + offset_left);
    }
  };

  // Creates a tick at the specified value at the position located in this.sliderTicks array
  DialogCtrlScrub.prototype.createTick = function(value, isSnapshot) {
    this.activeTicks[value] = true;
    // Choose image depending on whether there it is a user-generated snapshot
    // or a programmatically-generated keyframe
    var src = isSnapshot ? "img/snapshot.png" : "img/keyframe.png";
    var tick = $('<img src="' + src + '" />').addClass('tick-mark').attr("value", value);
    tick.css('left', this.sliderTicks[value] + '%').appendTo(ui.elems.$scrubberCont); 
  };

  // Creates the physical tick marks on the html page
  DialogCtrlScrub.prototype.addSliderTicks = function() {
    this.activeTicks = [];
    $("img.tick-mark").remove();
    // Generate tick marks for snapshots
    var snapshotsLen = BoardState.snapshots.length;
    for (var i = 0; i < snapshotsLen; i++) {
      var iVal = BoardState.snapshots[i];
      if (!this.activeTicks[iVal]) {
        this.createTick(iVal, true); //true -> snapshot ticks
      }
    }
    var keyframesLen = BoardState.keyframes.length;
    for (var j = 0; j < keyframesLen; j++) {
      var jVal = BoardState.keyframes[j];
      if (!this.activeTicks[jVal]) {
        this.createTick(jVal, false); //true -> snapshot ticks
      }
    }
  };

  // Seeks to the nearest snapshot tick in the given direction
  // -1 for backward, +1 for forward
  DialogCtrlScrub.prototype.seekNearestSnapshotTick = function(direction) {
    var current = dialogCtrl.board.messageIndex;
    var ticks = $('.tick-mark').map(function(_, el) {
      return $(el).attr("value");
    }).get();
    var smallestDiff = 999999;
    var nearestTick = -1;
    var ticksLen = ticks.length;
    for(var i = 0; i < ticksLen; i++) {
      var tickNum = parseInt(ticks[i]);
      var diff = Math.abs(tickNum - current);
      if (diff < smallestDiff) {
        if ((direction == -1 && tickNum < current) || (direction == 1 && tickNum > current)) {
          smallestDiff = diff;
          nearestTick = tickNum;
        }
      }
    }
    if (nearestTick != -1) {
      self.setScrubTime(nearestTick);
    } else {
      if (direction == 1) {
        self.setScrubTime(BoardState.messageCount);
      } else {
        self.setScrubTime(0);
      }
    }
  };

  // Called by add-on to jump to a specific message
  DialogCtrlScrub.prototype.seekTo = function() {
    var index = obj.eventData;
    this.setScrubTime(index);
  };

  // push step interval and increment step interval (acceleration)
  DialogCtrlScrub.prototype.pushAccelStepInterval = function(direction) {
    pushStepInterval(direction);
    incrementStepInterval(direction);
  };

  function pushStepInterval(direction) {
    if (direction === 1) {
      self.stepForwardInterval.push(setInterval(function(){ self.stepScrubTime(1); }, PLAYBACK_DELAY));
    } else if (direction === -1) {
      self.stepBackInterval.push(setInterval(function(){ self.stepScrubTime(-1); }, PLAYBACK_DELAY));
    }
  }

  function incrementStepInterval(direction) {
    if (direction === 1) {
      self.incrementForwardInterval.push(setInterval(function() { pushStepInterval(1); }, ACCELERATION_DELAY));
    } else if (direction === -1) {
      self.incrementBackInterval.push(setInterval(function() { pushStepInterval(-1); }, ACCELERATION_DELAY));
    }
  }  

  DialogCtrlScrub.prototype.playPause = function(pauseCallback, playCallback, restartCallback) {
    if (self.playing) {
      self.pause();
      pauseCallback();
    } else {
      if (parseInt(ui.elems.$scrubber.val()) != parseInt(ui.elems.$scrubber[0].max)) {
        self.play();
        playCallback();
      } else {
        self.restart();
        restartCallback();
      }
    } 
    dialogCtrl.board.updateLive();
  };

  DialogCtrlScrub.prototype.clearAccelStepIntervals = function() {
    var i;
    var stepBackIntervalLen = self.stepBackInterval.length;
    for(i = 0; i < stepBackIntervalLen; i++) {
      clearInterval(self.stepBackInterval[i]);
    }
    var stepForwardIntervalLen = self.stepForwardInterval.length;
    for(i = 0; i < stepForwardIntervalLen; i++) {
      clearInterval(self.stepForwardInterval[i]);
    }
    var incrementBackIntervalLen = self.incrementBackInterval.length;
    for(i = 0; i < incrementBackIntervalLen; i++) {
      clearInterval(self.incrementBackInterval[i]);
    }
    var incrementForwardIntervalLen = self.incrementForwardInterval.length;
    for(i = 0; i < incrementForwardIntervalLen; i++) {
      clearInterval(self.incrementForwardInterval[i]);
    }
    var stepTimeoutLen = self.stepTimeout.length;
    for(i = 0; i < stepTimeoutLen; i++) {
     clearTimeout(self.stepTimeout[i]);
    }
    self.stepBackInterval = [];
    self.stepForwardInterval = [];
    self.incrementBackInterval = [];
    self.incrementForwardInterval = [];
  };

  DialogCtrlScrub.prototype.startAcceleratedStepForward = function() {
    self.stepTimeout.push(setTimeout(function() {
      self.pushAccelStepInterval(1);
    }, PLAYBACK_START_DELAY));
  };

  DialogCtrlScrub.prototype.startAcceleratedStepBack = function() {
    self.stepTimeout.push(setTimeout(function() {
      self.pushAccelStepInterval(-1);
    }, PLAYBACK_START_DELAY));
  };
    

  return DialogCtrlScrub;
}();