/*
  ==================================================================================================
  _____ _   _ _  _______     _______ _    _  _____ _    _          _   _ _____  _      ______ _____  
 |_   _| \ | | |/ /  __ \ /\|__   __| |  | |/ ____| |  | |   /\   | \ | |  __ \| |    |  ____|  __ \ 
   | | |  \| | ' /| |__) /  \  | |  | |__| | (___ | |__| |  /  \  |  \| | |  | | |    | |__  | |__) |
   | | | . ` |  < |  ___/ /\ \ | |  |  __  |\___ \|  __  | / /\ \ | . ` | |  | | |    |  __| |  _  / 
  _| |_| |\  | . \| |  / ____ \| |  | |  | |____) | |  | |/ ____ \| |\  | |__| | |____| |____| | \ \ 
 |_____|_| \_|_|\_\_| /_/    \_\_|  |_|  |_|_____/|_|  |_/_/    \_\_| \_|_____/|______|______|_|  \_\
                                                                            
                                                                                                     
  A library for parsing messages into paths and drawing them onto canvases
  ==================================================================================================
*/
var InkPathsHandler = function() {
  'use strict';

  function InkPathsHandler() {}

  InkPathsHandler.prototype.drawScaledPathsBezier = function(message, context, scale, color) {
    var paths = message.paths;
    var startingIndex = 0;

    // Setting width of brush
    var diameter = BoardState.constants.nib_default;
    if (message.type == BoardState.constants.type_pen) {
      diameter = message.nibSize;
    }
    if (message.type == BoardState.constants.type_erase) {
      if (paths[0].type == BoardState.constants.path_circle) {
        diameter = paths[0].radius * 2;
        startingIndex = 1;
      } else {
        diameter = BoardState.constants.nib_eraser;
      }
    }

    // Setting color of brush
    var penColor = "#" + this.getHexColor(color.R) + this.getHexColor(color.G) + this.getHexColor(color.B);
    var ink = (message.type == BoardState.constants.type_pen) ? penColor : BoardState.constants.ink_eraser;
    context.strokeStyle = ink;

    //set a min threshold for stroke diameter
    var minLineWidth = 1;
    context.lineWidth = Math.max(diameter * scale, minLineWidth); 
    context.lineCap = 'round';
    context.lineJoin = 'round';

    var x = message.x;
    var y = message.y;
    var pathLen = paths == undefined ? 0 : paths.length;
    for (var j = startingIndex; j < pathLen; j++) {
      context.beginPath();
      context.moveTo(x * scale, y * scale);
      if(paths[j].type === "bez") {
        context.bezierCurveTo((x + paths[j].cpx1) * scale, (y + paths[j].cpy1) * scale, (x + paths[j].cpx2) * scale, (y + paths[j].cpy2) * scale, (x + paths[j].dx) * scale, (y + paths[j].dy) * scale);
      } else if(paths[j].type === "line") {
        context.lineTo((x + paths[j].dx) * scale, (y + paths[j].dy) * scale);
      }
      context.stroke();
      x = x + paths[j].dx;
      y = y + paths[j].dy;
    }
  };


  // Draws the line given by paths onto the given Canvas Context
  InkPathsHandler.prototype.drawScaledPaths = function(paths, context, scale, color) {
    // X and Y coordinates to draw lines between
    var x = paths.x;
    var y = paths.y;
    // Eraser or pen
    var penColor = "#" + this.getHexColor(color.R) + this.getHexColor(color.G) + this.getHexColor(color.B);
    var ink = (paths.type == BoardState.constants.type_pen) ? penColor : BoardState.constants.ink_eraser;
    context.save(); // Push state of context onto stack
    // Begin the brush stroke
    context.beginPath();
    context.strokeStyle = ink;
    //set a min threshold for stroke diameter
    var minLineWidth = 1; 
    context.lineWidth = Math.max(paths.diameter * scale, minLineWidth); 
    context.lineCap = 'round';
    context.lineJoin = 'round';
    context.moveTo(x[0]*scale, y[0]*scale);
    var xLen = x.length;
    for (var i = 1; i < xLen; i++) {
      context.lineTo(x[i]*scale, y[i]*scale);
    }
    context.stroke();
    // End brush stroke
    context.restore(); // Restore the previous state of the contexts
  };

  InkPathsHandler.prototype.getHexColor = function(num) {
    var hexVal = num.toString(16);
    if (hexVal.length == 1) {
      hexVal = hexVal + "0";
    }
    return hexVal;
  };

  // Converts the information in the JSON object message to paths to draw
  InkPathsHandler.prototype.JSONPathConverter = function(message, state) {
    var path = { x:[], y:[] };
    path.type = message.type;
    path.color = message.color;

    // Set the path size depending on nib and eraser
    path.diameter = state.constants.nib_default;
    if (message.type == state.constants.type_pen) {
      path.diameter = message.nibSize;
    }
    if (message.type == state.constants.type_erase) {
      if (message.paths[0].type == state.constants.path_circle) {
        path.diameter = message.paths[0].radius * 2;
      } else {
        path.diameter = state.constants.nib_eraser;
      }
    }

    path.seqID = message.seqID;
    path.time = message.time;

    var lastX = message.x;
    var lastY = message.y;
    var nextX = lastX;
    var nextY = lastY;
    this.addPoint(path, lastX, lastY);

    var pathsLen = message.paths.length;
    for (var i = 0; i < pathsLen; i++) {
      var subPath = message.paths[i];
      switch(subPath.type) {
        case state.constants.path_line:
          lastX += subPath.dx;
          lastY += subPath.dy;
          this.addPoint(path, lastX, lastY);
          break;
        case state.constants.path_curve:
          nextX += subPath.dx;
          nextY += subPath.dy;
          var subCurve = this.convertBezierCurvesToPointsTrajectory(
            lastX, lastY,               // start point
            lastX + subPath.cpx1, lastY + subPath.cpy1, // control point 1
            lastX + subPath.cpx2, lastY + subPath.cpy2, // control point 2
            nextX, nextY                // end point
          );
          this.addPoints(path, subCurve.x, subCurve.y);
          lastX = nextX;
          lastY = nextY;
          break;
        default:
          // Other stroke styles not yet implemented
          break;
      }
    }
    return path;
  };

  // Add a single point (x,y) to  obj
  InkPathsHandler.prototype.addPoint = function(obj, x, y) {
    obj.x.push(x);
    obj.y.push(y);
  };

  // Add an array of points to obj
  InkPathsHandler.prototype.addPoints = function(obj, x_array, y_array) {
    var xLen = x_array.length;
    for (var i = 0; i < xLen; i++) {
      obj.x.push(x_array[i]);
      obj.y.push(y_array[i]);
    }
  };

  // Algorithm adapted from
  // whiteboard/source/libRemoteInkService/InkCurveToPointsConverter.cpp
  // in https://hg.smarttech.com/EnterpriseNativeWhiteboard/whiteboard
  //
  // For an input cubic Bezier curve, return a set of points on the curve
  // B(t) = S(t^3) + 3Ct(1-t)^2 + 3P(1-t)t^2 + Et^3
  //   where
  //   S = start point
  //   C = control point 1
  //   P = control point 2
  //   E = end point
  //   t = {0..1}
  //
  // Number of points returned is "delta", which is configurable
  //
  // http://en.wikipedia.org/wiki/B%C3%A9zier_curve#Cubic_B.C3.A9zier_curves
  //
  //
  InkPathsHandler.prototype.convertBezierCurvesToPointsTrajectory = function(sx, sy, cpx1, cpy1, cpx2, cpy2, ex, ey) {
    var pointsTrajectory = { x:[], y:[] };
    var done = false;
    var s = this.getTrajectoryPointDistance();
    var t = s;
    var t_inv, x, y;
    var floatingPointCorrectionFactor = 10;

    while (!done) {
      if (t >= 1.0) {
        // plug in final point e for t=1, don't need
        // to run the math
        t = 1.0;
        pointsTrajectory.x.push(ex);
        pointsTrajectory.y.push(ey);
        done = true;
      } else {
        t_inv = 1.0 - t;
        x = Math.pow(t_inv, 3) * sx + 3 * t * Math.pow(t_inv, 2) * cpx1 + 3 * t_inv * Math.pow(t, 2) * cpx2 + Math.pow(t, 3) * ex;
        y = Math.pow(t_inv, 3) * sy + 3 * t * Math.pow(t_inv, 2) * cpy1 + 3 * t_inv * Math.pow(t, 2) * cpy2 + Math.pow(t, 3) * ey;

          pointsTrajectory.x.push(x);
          pointsTrajectory.y.push(y);

        // Note: repeated addition of floating point numbers will
        // cause this loop to eventually accumulate floating point
        // errors. For s=0.1, after 9 iterations t will be 0.9999999999999999.
        // We deal with this by using a correction factor of 1/s = 10
        // so that we are adding integers.
        t *= floatingPointCorrectionFactor;
        s *= floatingPointCorrectionFactor;
        t += s;
        t /= floatingPointCorrectionFactor;
        s /= floatingPointCorrectionFactor;
      }
    }
    
    return pointsTrajectory;
  };

  InkPathsHandler.prototype.getTrajectoryPointDistance = function() {
    var DELTA = 0.1;
    return DELTA;
  };

  // InkPathsHandler.prototype.setTrajectoryPointDistance = function(n) {
  //   DELTA = n;
  //   floatingPointCorrectionFactor = 1.0/n;
  // };
  
  return InkPathsHandler;
}();