/*
  ==================================================================================================
  _   _ ____ _____ _   _ _______ ______ _____  ______      _____ ______ 
 | \ | |  _ \_   _| \ | |__   __|  ____|  __ \|  ____/\   / ____|  ____|
 |  \| | |_) || | |  \| |  | |  | |__  | |__) | |__ /  \ | |    | |__   
 | . ` |  _ < | | | . ` |  | |  |  __| |  _  /|  __/ /\ \| |    |  __|  
 | |\  | |_) || |_| |\  |  | |  | |____| | \ \| | / ____ \ |____| |____ 
 |_| \_|____/_____|_| \_|  |_|  |______|_|  \_\_|/_/    \_\_____|______|
                                                                        
  
  Communication between the dialog and NB
  ==================================================================================================
*/
var NBInterface = function() {
  'use strict';
  
  var self;
  var inkPathsHandler;
  var ui;
  var dialogCtrl;

  function NBInterface(pInkPathsHandler, pUi) {
    self = this;
    // Set these again for scoping issues
    inkPathsHandler = pInkPathsHandler;
    ui = pUi;
    this.WAIT_INTERVAL = 10; //wait every WAIT_INTERVAL lines to give NB time to rest
  }

  // observerCallbacks: array of [stringToListenFor, callback] pairs
  // onReadyCallback: a callback function to be executed when NB is ready for communication
  NBInterface.prototype.onNbReady = function(observerCallbacks, onReadyCallback) {
    NB.ready(function() {
      // Setup observers
      var observerCallbacksLen = observerCallbacks.length;
      for (var i=0; i<observerCallbacksLen; i++) {
        var observerPair = observerCallbacks[i];
        window.NB.addObserver(observerPair[0], observerPair[1]);
      }
      // Execute ready callback
      onReadyCallback();
    });
  };

  NBInterface.prototype.getPageVisibleArea = function() {
    return NB.page.getVisibleArea();
  };

  NBInterface.prototype.addPage = function() {
    return NB.document.addPage();
  };

  NBInterface.prototype.postEvent = function(eventName, data) {
    if (data === undefined) {
      data = {};
    }
    var dataContainer = {};
    dataContainer['com.smarttech.KAPP'] = data;
    window.NB.postEvent(eventName, dataContainer);
  };

  // Called everytime we poll for new messages.
  // Keeps track of what sequence we're on.
  // Performs different calls according to message type.
  // Registers all draw calls into BoardState.messages
  NBInterface.prototype.parseMessages = function(data) {
    if (typeof data == 'undefined') {
      return;
    }
    var seq_id = BoardState.session.sequenceId;
    var dataLen = data.length;
    for (var i = 0; i < dataLen; i++) {
      seq_id++;
      if (seq_id != data[i].seq) { // Check to make sure the sequence is correct
        return; 
      }
      BoardState.session.sequenceId = seq_id; // So we know from where to fetch next
      // TODO: Check to see if message is already seen (not needed said Siamak)
      var messages = data[i].messages;
      var messageCount = messages.length;
      for (var j = 0; j < messageCount; j++) {
        var message = messages[j];
        if (typeof message == 'undefined') {
          continue;
        }
        if (message.color && (!message.color.R || !message.color.G || !message.color.B)) {
          message.color = this.parseHexColor(message.color);
        }
        if (message.type == BoardState.constants.type_pen || message.type == BoardState.constants.type_erase) {
          this.handleStroke(message);
        }
      }
    }
    BoardState.messages.sort(this.messageCompare);
  };

  // Convert hex color string e.g. "#FFFFFF" to JSON color object
  NBInterface.prototype.parseHexColor = function(hex) {
    //If we don't know how to handle the color encoding, assume black
    if (hex.length != 7) {
      return {R:0, G:0, B:0};
    }
    var color = {};
    color.R = parseInt(hex.substring(1, 3), 16);
    color.G = parseInt(hex.substring(3, 5), 16);
    color.B = parseInt(hex.substring(5, 7), 16);
    return color;
  };

  // Eraser or Pen stroke
  NBInterface.prototype.handleStroke = function(message) {
    BoardState.messages[BoardState.messageCount++] = message;
  };

  // Compares the time stamps of two messages
  NBInterface.prototype.messageCompare = function(a, b) {
    if (a.time < b.time) {
      return -1;
    } else if (a.time > b.time) {
      return 1;
    } else {
      return 0;
    }
  };

  // Draws the given data in NB as inkstrokes (converts eraser+ink strokes into just ink strokes)
  // options: crop: {width, height, x, y}, inkX, inkY, inkWidth, messageIndex, page
  NBInterface.prototype.inkToPage = function(data, options) {
    // Setup
    var messagesArray = data.messages;
    var width = data.session.boardWidth;
    var height = data.session.boardHeight;

    // Constructs empty JSON object for options if it is undefined so following checks work
    options = options ? options : {};

    // Setup for the color of the ink (default is black)
    var black = {R: 0, G: 0, B: 0};
    // Setup for the message to draw up to (default is current)
    var messageIndex = options.messageIndex;
    // Choose which page to draw to
    var page = options.page ? options.page : NB.page;
    // Setup for the location to draw ink to in Notebook
    var xOffset = options.inkX ? options.inkX : 0;
    var yOffset = options.inkY ? options.inkY : 0;
    var NB_width = options.inkWidth ? options.inkWidth : NB.page.width;

    // Scale for scaling down both the eraser canvas and the actual ink drawn in NB
    // to the width of NB page. This scaling is important so that the snapshot
    // can run quickly, even for extremely large kapp boards. 
    var scale = 1;
    if (options.inkWidth) {
      scale = NB_width / width;
    } else {
      if (width / NB_width > height / NB.page.height) {
        scale = NB_width / width;
      } else {
        scale = NB.page.height / height;
      }
    }

    // Setup for cropping if necessary (defaults to "crop" to full board width and height)
    //var crop = options.crop ? options.crop : {width: width * scale, height: height * scale, x: 0, y: 0};
    var crop = {};
    if (options.crop) {
      crop.width = options.crop.width * scale;
      crop.height = options.crop.height * scale;
      crop.x = options.crop.x * scale;
      crop.y = options.crop.y * scale;
    } else {
      crop = {width: width * scale, height: height * scale, x: 0, y: 0};
    }

    var erase_canvas = document.createElement("canvas");
    var erase_context = erase_canvas.getContext("2d");
    erase_context.canvas.width = width * scale;
    erase_context.canvas.height = height * scale;

    // Ink to be added to NB.
    var linesGroup = [];
    // Traverse array in reverse order allows us to do the ink and eraser strokes
    // easily. Consider the following reasoning to see why we need to go in reverse order:
    /*
    Each erase stroke affects all messages before it, but none after it. 
    Consider the following simple case:
    Start session
    ...
    Draw
    Take snapshot

    Then, notice that the last draw can be put on the board without thinking because 
    anything drawn/erased beforehand don't impact it.
    
    Following that logic, if we draw the most recent messages (the ones near 
    messagesArray[messagesArray.length - 1]) first, then you can see that eraser
    strokes that are drawn will only impact the ink strokes that we will draw 
    later on and therefore will allow us to check if an area that an ink stroke
    is about to be drawn in had been erased or not.
    */
    for (var i = messageIndex - 1; i >= 0; i--) {
      // Convert the json data into a path to draw
      var paths = inkPathsHandler.JSONPathConverter(messagesArray[i], data);

      // Drawing ink
      if (paths.type == data.constants.type_pen) { 
        // X and Y coordinates to draw lines between
        var x = paths.x;
        var y = paths.y;
        // Initialize the data about the previous point in the message
        var currColor = messagesArray[i].color ? messagesArray[i].color : black;
        var prevInfo = this.startNewLine(paths, 0, scale, erase_context, xOffset, yOffset, currColor, crop);
        var scaledX, scaledY, pixelColor, onEraser;

        /*
        Loop invariant: prevInfo.x, .y, .onEraser contain the respective values for 
        the previous point in the message 
        */
        var xLen = x.length;
        for (var j = 1; j < xLen; j++) {
          // Check if the pixel that the point of ink is in is also in a pixel
          // where an erase stroke is
          scaledX = x[j] * scale;
          scaledY = y[j] * scale;
          onEraser = this.checkOnEraser(scaledX, scaledY, erase_context, crop);
          // If a point is inside an erase stroke, find the closest point between this point
          // and the previous point of the message (that is outside of an eraser stroke).
          if (onEraser) {
            if (!prevInfo.onEraser) {
              var xy = this.bSearch(scaledX, scaledY, prevInfo.x, prevInfo.y, erase_context, crop);
              prevInfo.line.addPoint(xy[0] + xOffset, xy[1] + yOffset);
              linesGroup.push(prevInfo.line);
              j++;
              if (j < xLen) {
                // Since we are making a new line, need to reinitialize the data about
                // the previous point in the message
                currColor = messagesArray[i].color ? messagesArray[i].color : black;
                prevInfo = this.startNewLine(paths, j, scale, erase_context, xOffset, yOffset, currColor, crop);
              }
            } else {
              prevInfo.x = scaledX;
              prevInfo.y = scaledY;
              prevInfo.onEraser = onEraser;
            }
          // If a point is not inside an eraser stroke, then it will be added to the line.
          } else {
            // If the previous point was inside an eraser stroke, then put in an extra point
            if (prevInfo.onEraser) {
              var xy = this.bSearch(prevInfo.x, prevInfo.y, scaledX, scaledY, erase_context, crop);
              prevInfo.line.addPoint(xy[0] + xOffset, xy[1] + yOffset); 
            }
            prevInfo.line.addPoint(scaledX + xOffset, scaledY + yOffset);
            prevInfo.x = scaledX;
            prevInfo.y = scaledY;
            prevInfo.onEraser = onEraser;
          }
        }
        if (prevInfo.line.points !== "") {
          linesGroup.push(prevInfo.line);
        }
      // Draw eraser stroke on eraser canvas
      } else { 
        inkPathsHandler.drawScaledPaths(paths, erase_context, scale, {R: 255, G: 255, B: 255});
      }
    }
    // Array to hold to line objects (i.e. lines that have already been inserted into the NB page)
    var lineObjsGroup = [];
    // Draw the line objects in order sorted by timestamp (we have to go through the array
    // in reverse because the lines were pushed onto the array in reverse order)
    this.addLineObjects(linesGroup.length - 1, linesGroup, lineObjsGroup, page);
  };

  NBInterface.prototype.addLineObjects = function(i, linesGroup, lineObjsGroup, page) {
    var linesGroupLen = linesGroup.length;
    var fracCompleted = (linesGroupLen - 1 - i) / linesGroupLen;
    ui.status.loadingBarUpdate(fracCompleted);
    window.setTimeout(function() {
      self.loopBodyAddLineObjects(i, linesGroup, lineObjsGroup, page);
    }, 0);
  };

  // linesGroup: ink prototypes, lineObjsGroup: ink objects that have already been added to NB page
  NBInterface.prototype.loopBodyAddLineObjects = function(i, linesGroup, lineObjsGroup, page) {
    var end = Math.max(i - this.WAIT_INTERVAL, -1);
    for (i; i > end; i--) {
      var thickness = linesGroup[i].stroke.thickness;
      lineObjsGroup.push(page.addObject(linesGroup[i]));
      lineObjsGroup[lineObjsGroup.length-1].stroke.thickness = thickness;
    }
    if (i > 0) {
      this.addLineObjects(i, linesGroup, lineObjsGroup, page);
    } else {
      var eventData = {};
      if (lineObjsGroup.length > 1) {
        // Group lines in NB
        var groupedInk = page.groupObjects(lineObjsGroup);
        eventData["com.smarttech.KAPP"] = groupedInk.guid;
        NB.postEvent("addNewGroupId", eventData);
      }
      eventData["com.smarttech.KAPP"] = {};
      window.NB.postEvent("dialogFinishedDraw", eventData);
    }
  };

  // Binary searches between the eraseX,Y and inkX,Y and returns the point
  // on the line between them that is closest to eraseX,Y that has is not erased on the erase_context.
  // Pre: eraseX,Y is a pixel on the erase_context that is erased (color=[255,255,255])
  // and inkX,Y is a pixel on the erase_context that is not erased (color=different)
  NBInterface.prototype.bSearch = function(eraseX, eraseY, inkX, inkY, erase_context, crop) {
    // printDebug("Start of bSearch with eraseX,Y: " + eraseX + "," + eraseY + " and inkX,Y: " 
    //     + inkX + "," + inkY);
    var threshold = 1; //how accurate the bSearch should be (smaller # = more accurate)
    var midX = (eraseX + inkX) / 2;
    var midY = (eraseY + inkY) / 2; 
    var onEraser = this.checkOnEraser(midX, midY, erase_context, crop);

    // Done when we are not in an eraser stroke and the midX is within threshold pixels 
    // in both X and Y directions of the target point (since these are floating point numbers)
    while (onEraser || !((Math.abs(midX-eraseX) < threshold && Math.abs(midY-eraseY) < threshold) || 
        (Math.abs(midX-inkX) < threshold && Math.abs(midY-inkY) < threshold)) ) {
      // printDebug("midX,Y: " + midX + "," + midY + "; " + "inkX,Y: " + inkX + "," + inkY + "; eraseX,Y: " + eraseX + "," + eraseY + "; " + "onEraser: "+ onEraser);
      if (onEraser) {
        eraseX = midX; 
        eraseY = midY;
      } else {
        inkX = midX;
        inkY = midY;
      }
      midX = (eraseX + inkX) / 2;
      midY = (eraseY + inkY) / 2; 
      onEraser = this.checkOnEraser(midX, midY, erase_context, crop);
    }
    // printDebug("CHOSEN midX,Y: " + midX + "," + midY);
    return [midX,midY];
  };

  // Creates a new line, and does the initialization/fencepost step to draw the first point
  // of the line as well as save the scaledX, scaledY, onEraser, and line so that the next
  // point in the line can refer back to it.
  // Returns the prevInfo object corresponding to the values for the point at paths.x[index], paths.y[index]:
  // {prevX:#, prevY:#, prevOnEraser:T/F, line:Obj}
  NBInterface.prototype.startNewLine = function(paths, index, scale, erase_context, xOffset, yOffset, color, crop) {
    var line = NB.objectPrototype.ink();
    line.stroke.thickness = Math.max(paths.diameter * scale, 1);
    line.stroke.color = [color.R, color.G, color.B];
    var scaledX = paths.x[index] * scale;
    var scaledY = paths.y[index] * scale;
    var onEraser = this.checkOnEraser(scaledX, scaledY, erase_context, crop);
    if (!onEraser) {
      line.addPoint(scaledX + xOffset, scaledY + yOffset);
    }
    var prevInfo = {
      x: scaledX,
      y: scaledY,
      onEraser: onEraser,
      line: line
    };
    return prevInfo;
  };

  // Returns true iff the given x, y point is in a part of the erase_context that has been erased
  // or if it is outside of the cropping window. Essentially, tells whether an ink stroke
  // should be drawn or not based on the fact that it should only be drawn if it is both
  // inside the area to crop to and is not inside an eraser stroke. Note that if a point is outside
  // of a cropBox, then it is equivalent to having been erased, for all intents and purposes.
  NBInterface.prototype.checkOnEraser = function(scaledX, scaledY, erase_context, crop) {
    var pixelColor = erase_context.getImageData(scaledX, scaledY, 1, 1).data;
    var onEraser = pixelColor[0] == 255 && pixelColor[1] == 255 && pixelColor[2] == 255;
    var inCropBox = scaledX > crop.x && scaledX < crop.x + crop.width && scaledY > crop.y && scaledY < crop.y + crop.height;
    return onEraser || !inCropBox;
  };
  
  return NBInterface;
}();