// Handles storing, parsing, and sending out the message data from the kapp board

var SessionHandler = function() {

	var POLLING_RATE = 3000;
	var KEYFRAME_THRESHOLD = 3;

	var that;

	var BoardState;
	var kappHandler;
	var nbMessenger;

	// temporary buffer of snapshots stored by sequence ID
	var snapshots;

	var getMessagesInterval;

	function SessionHandler(kapp, messenger) {
		that = this;
		kappHandler = kapp;
		nbMessenger = messenger;
		snapshots = [];
		BoardState = getInitialState();
	}

	// Connect to the given sessionid
	SessionHandler.prototype.connect = function (sessionid, updateViews) {
		BoardState.status = "connecting";
		updateViews("connecting");
		var connectSuccess = function(response) {
			onConnect(response, updateViews);
		};
		var connectFailure = function(response) {
			BoardState.status = "error";
			var errcode = response.status;
      		updateViews("error", errcode);
		};
		kappHandler.connect(sessionid, connectSuccess, connectFailure);
	};

	// Disconnect from the board
	SessionHandler.prototype.disconnect = function (updateViews, errcode) {
		BoardState.status = "disconnected";
		// Stops the polling of new messages.
		clearInterval(getMessagesInterval);
		// Connection Status
		trackEvent('backend', 'messages', 'messages at disconnect', BoardState.messageCount);
		var sessionLength = (new Date()).getTime() - BoardState.connectTime;
		trackEvent('backend', 'connection', 'session length (ms)', sessionLength);
		BoardState.status = "disconnected";
		nbMessenger.disconnectDialog();
		if(errcode) {
			updateViews("error", errcode);
		} else {
			updateViews("disconnected");
		}
		printDebug("Disconnected");
	};

	// Open a dialog window
	SessionHandler.prototype.openDialog = function (updateViews) {
		var onOpenFunc = function() {updateViews("dialogOpened");};
		var onCloseFunc = function() {updateViews("dialogClosed");};
		nbMessenger.openDialog(BoardState, onOpenFunc, onCloseFunc);
	};

	// Check whether the session status matches the passed in value
	SessionHandler.prototype.checkSessionStatus = function(status) {
		return status == BoardState.status;
	};

	// Gets and handles messages once we've successfully connected to a board
	function onConnect(response, updateViews) {

		// Callback for when getting messages succeeds
		var getMessagesSuccess = function(response) {
			if (response) {
				printDebug("New messages received.");
				onMessagesReceived(response);
			}
		};

		// Callback for when getting messages fails
		var getMessagesFailure = function(response) {
			var errcode = response.status;
			// If we get a 404 while we are connected, this is a normal d/c with session expired
			if (errcode == 404) {
				printDebug("Session expired.");
				that.disconnect(updateViews);
			} else {
				printDebug("Failed to retrieve messages.");
				// If it's another error, this is not a normal d/c
				// Update Connection Status on failure
				that.disconnect(updateViews, errcode);
			}
		};

		// Success callback for the initial message retrieval
		var initialGetSuccess = function(response) {
			BoardState.status = "connected";
			getMessagesSuccess(response);
			getMessagesInterval = setInterval(function() {
				kappHandler.getMessages(BoardState.session.id, BoardState.session.auth, BoardState.session.sequenceId, getMessagesSuccess, getMessagesFailure);
			}, POLLING_RATE);
			updateViews("connected");
		};

		// Connecting to a session will always require us to use the new authToken
		BoardState.session.auth = response.authToken;
		BoardState.connectTime = (new Date()).getTime();
		if (response.sessionId != BoardState.session.id) {
			// Initialize the session state
			BoardState.session.id = response.sessionId;
			BoardState.session.boardWidth = response.kappBoard.width;    // Width of the kapp board drawing area
			BoardState.session.boardHeight = response.kappBoard.height;  // Height of the kapp board drawing area
			resetBoardState();
		}

		// Initial call to getMessages
		kappHandler.getMessages(BoardState.session.id, BoardState.session.auth, BoardState.session.sequenceId, initialGetSuccess, getMessagesFailure);
	}

	// Resets the messages, snapshots, etc stored in the Global State
	// Used when connecting to a new session
	function resetBoardState() {
		BoardState.messages = [];
		BoardState.messageCount = 0;
		BoardState.snapshots = [];
		BoardState.snapshotCount = 0;
		BoardState.keyframes = [];
		BoardState.keyframeCount = 0;
		BoardState.colorData = [];
		BoardState.session.sequenceId = 1; // First sequence requests seqid 1
		snapshots = [];
		nbMessenger.clearSessionData();
	}

	// Handles messages received from the board
	function onMessagesReceived(response) {
		// Get previous counts of messages
		var prevMessageCount = BoardState.messageCount;
		var prevSnapshotCount = BoardState.snapshotCount;
		// Parse the incoming messages to be ready for drawing
		parseMessages(response);
		// If we have new messages or this is the first time, update the dialog window
		var newMessages = BoardState.messageCount > prevMessageCount;
		var newSnapshots = BoardState.snapshotCount > prevSnapshotCount;
		if(newMessages || newSnapshots) { 
			nbMessenger.updateDialog(BoardState);
		}
	}

	// Called everytime we poll for new messages.
	// Keeps track of what sequence we're on.
	// Performs different calls according to message type.
	// Registers all draw calls into BoardState.messages
	function parseMessages(data) {
	  if (typeof data == 'undefined') { printDebug("Messages Received: undefined"); return; }
	  // Process each message
	  var seq_id = BoardState.session.sequenceId;
	  for (var i = 0; i < data.length; i++) {
	    if (seq_id++ != data[i].seq) { printDebug("Parse Messages: message out of order"); return; } // Check to make sure the sequence is correct 
	    BoardState.session.sequenceId = seq_id; // So we know from where to fetch next
	    // TODO: Check to see if message is already seen (not needed said Siamak)
	    var messages = data[i].messages;
	    for (var j = 0; j < messages.length; j++) {
	      var message = messages[j];
	      // Need to parse color if set in hex format e.g. #FFFFFF to RGB
	      if (message.color && (!message.color.R || !message.color.G || !message.color.B)) {
	        // iOS kapp app adds black coloring to all ink strokes - this clears it so time based color will work the same across iOS and Android
	        if(message.color === "#000000") {
	          message.color = undefined;
	        } else {
	          message.color = parseHexColor(message.color);
	        }
	      // Color was set previously, need to restore it
	      } else if (BoardState.colorData[BoardState.messageCount]) {
	        message.color = BoardState.colorData[BoardState.messageCount];
	      }
	      if (typeof message == 'undefined') { printDebug("Parse Messages: message undefined"); continue; }
	      switch (message.type) {
	        case BoardState.constants.type_page:
	          handleSnapshot(message);
	          break;
	        case BoardState.constants.type_pen:
	          handleStroke(message);
	          break;
	        case BoardState.constants.type_erase:
	          handleStroke(message);
	          break;
	        case BoardState.constants.type_clear:
	          clearBoard();
	          break;
	        default:
	          printDebug("Unknown Message Type: " + message.type);
	          break;
	      }
	    }
	  }
	  // Post process messages
	  BoardState.messages.sort(messageCompare);  // Sort the messages by time
	  indexSnapshots(); // Update the snapshots array with indices into the Message Array
	  generateKeyframes();  // Create keyframes based on eraser strokes
	}

	// Snapshot command - Type 0
	function handleSnapshot(message) {
		snapshots.push(message.seqID);
		timeSinceMessage = -1;
	}

	// Eraser or Pen stroke - Type 1 and 2
	function handleStroke(message) {
		// Save the stroke message
		BoardState.messages[BoardState.messageCount++] = message;
	}

	// Clear board - Type 3
	function clearBoard() {
		printDebug("Clear Board");
		BoardState.colorData = [];
		BoardState.messages = [];
		BoardState.messageCount = 0;
		BoardState.snapshots = [];
		BoardState.snapshotCount = 0;
		BoardState.keyframes = [];
		BoardState.keyframeCount = 0;

		nbMessenger.clearDialog(BoardState);
	}

	// Compares messages by timestamp ascending order
	function messageCompare(a, b) {
		if (a.time < b.time) {
			return -1;
		}
		if (a.time > b.time) {
			return 1;
		} else {
			return 0;
		}
	}

	// Convert hex color string e.g. "#FFFFFF" to JSON color object
	function parseHexColor(hex) {
		//If we don't know how to handle the color encoding, assume black
		if (hex.length != 7) {
			return {R:0, G:0, B:0};
		}
		var color = {};
		color.R = parseInt(hex.substring(1, 3), 16);
		color.G = parseInt(hex.substring(3, 5), 16);
		color.B = parseInt(hex.substring(5, 7), 16);
		return color;
	}

	// Stores snapshots as indices into the message array.
	// Stores into BoardState.snapshots
	function indexSnapshots() {
		BoardState.snapshots = [];
		snapshots.sort(numberComparator);
		for (var i = 0; i < BoardState.messages.length; i++) {
			var seqID = BoardState.messages[i].seqID;
			var snapshotID = snapshots[BoardState.snapshots.length];
			if (!snapshotID) break;
			// If the time on the snapshot is less than the message time, save it out
			// Also save it out if we're at the end and there's still a snapshot left
			if (snapshotID < seqID) {
				BoardState.snapshots.push(Math.max(0, i));
			} else if (i == BoardState.messages.length - 1) {
				BoardState.snapshots.push(BoardState.messages.length);
			}
		}
		BoardState.snapshotCount = BoardState.snapshots.length;
	}

	// Simple number comparator
	function numberComparator(a, b) {
		return a - b;
	}

	// Create Keyframes based on eraser strokes.
	// Stores into BoardState.keyframes
	function generateKeyframes() {
		var keyframes = [];
		var keyframeCount = 0;
		var eraserStrokes = 0;
		var messages = BoardState.messages;
		var length = messages.length;
		for(var i = 0; i < length; i++) {
			if(eraserStrokes != -1 && messages[i].type == BoardState.constants.type_erase) {
				eraserStrokes++;
				if(eraserStrokes >= KEYFRAME_THRESHOLD) {
					keyframes[keyframeCount++] = i - (eraserStrokes - 1);
					eraserStrokes = -1;
				}
			} else if(messages[i].type == BoardState.constants.type_pen) {
				eraserStrokes = 0;
			}
		}
		BoardState.keyframes = keyframes;
		BoardState.keyframeCount = keyframeCount;
	}

	// Returns a copy of a cleared, initial board state object
	function getInitialState() {
		return({
			"session": {
				"name": "",
				"password": "",
				"id": "",
				"auth": "",
				"sequenceId": "",
				"boardWidth": "",
				"boardHeight": ""
			},
			"constants": {
				// Message type identifiers
				"type_page": 0,
				"type_erase": 1,
				"type_pen": 2,
				"type_clear": 3,
				// Path type identifiers
				"path_line": "line",
				"path_curve": "bez",
				"path_circle": "cir",
				// Kapp board constants
				"ink_default": "#000000", // Default Ink Color
				"ink_eraser": "#FFFFFF", // Default eraser color
				"nib_default": 4,     // Default Nib Size (0.5mm increments)
				"nib_eraser": 560,    // Default eraser nib size
				"nib_factor": 5       // Conversion factor from Nib units to board units
			},
			"colorData": [],
			"messages": [],
			"messageCount": 0,
			"snapshots": [],
			"snapshotCount": 0,
			"keyframes": [],
			"keyframeCount": 0,
			"connectTime" : 0,
			"status": "disconnected"
		});
	}


	return SessionHandler;

}();